var exec = require('child_process').exec,
  async = require('async'),
  path = require('path'),
  fs = require('fs'),
  bplistParse = require('bplist-parser'),
  Tail = require('tail').Tail,
  spawn = require('child_process'),
  _ = require("underscore");

var IOS_SIM_VERSION = '0.1';

var simulatorAppId = "com.apple.iphonesimulator",
  deviceProperty = "SimulateDevice",
  deviceIphoneRetina3_5InchiOS7 = "iPhone Retina (3.5-inch)",
  deviceIphoneRetina4_0InchiOS7 = "iPhone Retina (4-inch)",
  deviceiPhoneRetine4_0InchiOS764bit = "iPhone Retina (4-inch 64-bit)",
  deviceiPadRetinaiOS764bit = "iPad Retina (64-bit)",
  deviceIpadRetinaiOS7 = "iPad Retina",
  deviceIphone = "iPhone",
  deviceIpad = "iPad",
  sdkRoot = null,
  stdoutFileHandle,
  stderrFileHandle,
  exitOnStartup = false,
  retinaDevice = false,
  tallDevice = false,
  sim_64bit = false,
  startOnly,
  verbose,
  alreadyPrintedData = false,
  launchFlag,
  CFBundleIdentifier=null,
  isXcode6;


// Finds the developer dir via xcode-select or the DEVELOPER_DIR environment
// variable.
function FindDeveloperDir(callback) 
{
  // Go look for it via xcode-select.
  exec("/usr/bin/xcode-select -print-path", function(error, stdout, stderr){
    var output = stdout.replace(/(\r\n|\n|\r)/gm,"");
    //console.log("[DEBUG] developerDir:",output);
    if(output.length==0) output = null;
    callback(output);
  });
}

function showSDKs() 
{
  exec("xcodebuild -showsdks", function(error, stdout, stderr){
    var output = stdout.replace(/(\r\n|\n|\r)/gm,"");
    console.log("Available iOS SDK's");
    output.split('iOS Simulator SDKs:\t')[1].split('\t').forEach(function(val){
      if(val.indexOf('Simulator - ')>-1)console.log("[DEBUG]",val.split('Simulator - ')[1]);
    });
  });
}

function showAllSimulators() 
{
  //exec("Xcodebuild -sdk -version", function(error, stdout, stderr){
  getAllDevices(function(output){
    console.log("Available Simulators");
    console.log(output);
  });
}

function getAllDevices(callback)
{
  exec("xcrun simctl list devices", function(error, stdout, stderr){
    if (error) {
      console.error('Could not list devices', error); 
      return;
    }
    var deviceSecRe = /-- iOS (.+) --(\n .+)*/mg;
    var matches = [];
    var devices = {};
    var match = deviceSecRe.exec(stdout);

    while (match !== null) {
      matches.push(match);
      match = deviceSecRe.exec(stdout);
    }

    if (matches.length < 1) {
      console.error('Could not find device section', error); 
      return;
    }

    _.each(matches, function (match) {
      var sdk = match[1];
      devices[sdk] = [];
      _.each(match[0].split("\n").slice(1), function (line) {
        var lineRe = /^ (.+) \((.+)\) \((.+)\)/;
        var match = lineRe.exec(line);

        if (match === null) {
          console.error('Could not match line', error); 
          return;
        }
        
        var device = {};
        device.name = match[1];
        device.udid = match[2];
        device.state = match[3];
        devices[sdk].push(device);
      });
    });
    callback(devices);
  });
}

function launchSimulator(udid,callback)
{
  getAllDevices(function(devices) {
    var deviceToLaunch = null;
    Object.keys(devices).forEach(function(sdk){
      devices[sdk].forEach(function(simulatorDevice) {
        if(simulatorDevice.udid!=udid && simulatorDevice.state=='Booted'){
          // shutdown simulator
          console.error("[ERROR]","Implement simulator shutdown");
          process.exit(1);
        }
        if (simulatorDevice.udid==udid) deviceToLaunch = simulatorDevice;
      });
    });

    exec("open -a 'iOS Simulator' --args -CurrentDeviceUDID "+deviceToLaunch.udid, function(error, stdout, stderr){
      if(verbose) {
        console.log("[DEBUG]", 'Launching Simulator');
      }
      callback(stdout);
    });
  });
}

function launchAppInSimulator(udid, CFBundleIdentifier, callback)
{
  //exec
  exec("xcrun simctl launch "+udid+" "+CFBundleIdentifier, function(error, stdout, stderr){
    
    if(error==null) {
      var pid = stdout.split(' ')[1].split('\n')[0];
      if(verbose) {
        console.log("[DEBUG]",'Launching App');
        console.log("[DEBUG]",'pid',pid);
      }
      callback(pid);
    } else if(error.code==4) {
      console.error("[ERROR]", "Application not found on device");
    } else if(error.code == '146'){
      console.log("[NOTICE]", "Trying to launch app in 3s");
      setTimeout(function(){
        launchAppInSimulator(udid, CFBundleIdentifier, callback);
      },3000);
    } else if(error.code == '145') {
      console.error("[ERROR]", "Device not found");
    } else {
      console.error("[ERROR]", error);
    }
  });
}

function installApp(appPath, udid, callback)
{
  exec("xcrun simctl install "+udid+" "+appPath, function(err, stdout, stderr){
    console.log("[INFO]", "Trying to install App");
    if(!err) {
      console.log("[INFO]", "App Installed");
      callback();
    } else if(err.code == '146'){ //Invalid Device State
      setTimeout(function(){
        installApp(appPath, udid, callback);
      },3000);      
    } else if(verbose) {
      console.log("[INFO]", err);
    }
  });
}

function getPlistData(file) 
{
  var data;
  if (fs.existsSync(file)) {
    var fileData = fs.readFileSync(file);
    try {
      data = bplistParse.parseBuffer(fileData)[0];
    } catch (err) {
      if (err.message.indexOf("Invalid binary plist") !== -1) {
        data = xmlplist(file)[0];
      } else {
        throw err;
      }
    }
  } else {
    console.error("Settings file " + file + " did not exist");
    process.exit(1);
  }
  return data;
}

function resolvePath (string) 
{
  if (string.substr(0,1) === '~') {
    string = process.env.HOME + string.substr(1);
  }
  return path.resolve(string);
}

function launchApp(appPath, family, timeout, udid, uuid, environment, stdoutPath, stderrPath, args)
{
  //console.log(path, family, timeout, udid, uuid, environment, stdoutPath, stderrPath, args);
  async.series([
    function(next) {
      //check if simulator is running
      //xcrun simctl
      // Launch Simulator
      launchSimulator(udid,next);
    },
    function(next) {
      // Get appPath plist
      if(CFBundleIdentifier==null) {
        //install app
        installApp(appPath, udid, next);

        var plistData = getPlistData(appPath+'/info.plist');
        CFBundleIdentifier = plistData.CFBundleIdentifier;
      } else {
        next();
      }
      //console.log(CFBundleIdentifier);
    }, function(next) {
      var appName = path.basename(appPath);
     // console.log(appName);
      //launch app
      //check if app is running
      launchAppInSimulator(udid,CFBundleIdentifier,function(pid){
        //Get logs
        console.log("test");
        var logPath = resolvePath("~/Library/Logs/CoreSimulator/"+udid+"/system.log");
        console.log(logPath);
        tail = new Tail(logPath);
        tail.on("line", function(data) {
          //testing stuff
          console.log('[LOG]', data.trim());
        });

      });
    }
  ]);
}

function runWithArgv(argv) 
{
  argv.splice(0,1);
  argc = argv.length;

  if (argc < 2) {
    //[self printUsage];
    return;
  }

  /* Initializing variables*/
  var timeout = 90,
    developerDir = null,
    appPath = null;

  launchFlag = argv[1] == 'launch';
  startOnly = argv[1] == 'start';

  async.series([
    function(next) {
      // Get Developer Directory
      FindDeveloperDir(function(dir){
        developerDir = dir;
        if (developerDir == null) {
          console.error("Unable to find developer directory.");
          return;
        }
        next();
      });
    },
    function(next) {

    }
  ]);

  var numOfArgs;
  if (startOnly) {
    numOfArgs = 2;
  } else if (argc > 2) {
    if(argv[2] == '--cfbundleidentifier' ) {
      CFBundleIdentifier = argv[3];
      appPath=null;
      numOfArgs=4;
    } else {
      numOfArgs = 3;
      //appPath = path.dirname(argv[2]);
      appPath = path.resolve(argv[2]);
      console.log(appPath);
    }

  }
  
  var i = numOfArgs;

  if (argv[1] == "showsdks") {
    showSDKs();
    return;
  } else if (argv[1] == "showallsimulators") {
    /*for (i = 2; i < argc; i++) {
      if (argv[i] == "--xcode-dir") {
        if (++i < argc) {
          developerDir = argv[i];
        }
      }
    }*/
    showAllSimulators();
    return;
  } else if (launchFlag || startOnly) {
    if (launchFlag && argc < 3) {
      console.log("Missing application path argument");
      //[self printUsage];
      return;
    }

    var family = null,
      uuid = null,
      stdoutPath = null,
      stderrPath = null,
      udid = null,
      environment = {};
    for (; i < argc; i++) {
      if (argv[i] == "--version") {
        console.log(IOS_SIM_VERSION);
        return;
      } else if (argv[i] == "--help") {
        //[self printUsage];
        return;
      } else if (argv[i] == "--verbose") {
        verbose = true;
      } else if (argv[i] == "--exit") {
        exitOnStartup = true;
      } else if (argv[i] == "--sdk") {
        i++;
        var ver = argv[i]
        //'xcrun --sdk '+ver+' SDKROOT'
        if (sdkRoot == null) {
          console.log("Unknown or unsupported SDK version: ",argv[i]);
          showSDKs();
          return;
        }
      } else if (argv[i] == "--family") {
        i++;
        family = argv[i];
      } else if (argv[i] == "--uuid") {
        i++;
        uuid = argv[i];
      } else if (argv[i] == "--setenv") {
        i++;
        NSArray *parts = argv[i].split('=');
        environment.parts[1] = parts[0]
      } else if (argv[i] == "--env") {
        i++;
        /*NSString *envFilePath = [[NSString stringWithUTF8String:argv[i]] expandPath];
        environment = [NSMutableDictionary dictionaryWithContentsOfFile:envFilePath];*/
        if (!environment) {
          console.log("Could not read environment from file:", argv[i]);
          //[self printUsage];
          return;
        }
      } else if (argv[i] == "--stdout") {
        i++;
        //stdoutPath = [[NSString stringWithUTF8String:argv[i]] expandPath];
        console.log("stdoutPath:", stdoutPath);
      } else if (argv[i] == "--stderr") {
        i++;
        //stderrPath = [[NSString stringWithUTF8String:argv[i]] expandPath];
        console.log("stderrPath: %@", stderrPath);
      } else if (argv[i] == "--args") {
        i++;
        break;
      } else if (argv[i] == "--retina") {
        retinaDevice = YES;
      } else if (argv[i] == "--tall") {
        tallDevice = YES;
      } else if (argv[i] == "--xcode-dir") {
        i++;
        developerDir = argv[i];
      } else if (argv[i] == "--sim-64bit") {
        sim_64bit = YES;
      } else if (argv[i] == "--timeout"){
        i++;
        timeout = argv[i];
      } else if (argv[i] == "--udid"){
        i++;
        udid = argv[i];
      } else {
        console.log("unrecognized argument:", argv[i]);
        //[self printUsage];
        return;
      }
    }
    i = Math.min(argc, i);
    var args = [];
    for (; i < argc; i++) {
      args.push(argv[i]);
    }

    if (sdkRoot == null) {
      //[self LoadSimulatorFramework:developerDir];
      //Class systemRootClass = [self FindClassByName:@"DTiPhoneSimulatorSystemRoot"];
      //sdkRoot = [systemRootClass defaultRoot];
    }

    /* Don't exit, adds to runloop */
    launchApp(appPath, family, timeout, udid, uuid, environment, stdoutPath, stderrPath, args);
  } else {
    if (argc == 2 && argv[1] == "--help") {
      //[self printUsage];
      return;
    } else if (argc == 2 && argv[1] == "--version") {
      console.log(IOS_SIM_VERSION);
      return;
    } else {
      console.log("Unknown command");
      //[self printUsage];
      return;
    }
  }
}
runWithArgv(process.argv);
//node simctl.js  launch --cfbundleidentifier com.appcelerator.TestApp  --udid CE514312-2815-489A-BF44-700FCC89FE93 --verbose
//node simctl.js  launch build/Debug-iphonesimulator/TestApp.app --udid CE514312-2815-489A-BF44-700FCC89FE93 --verbose